/* Dyadic grids in arbitrary dimension */
/* Last edited on 2005-02-14 23:16:10 by stolfi */

#ifndef dygrid_H
#define dygrid_H

#include <stdlib.h>

typedef unsigned char nat8;
typedef unsigned long nat64;

/* NODE INDICES */

typedef nat64 dyg_node_index;
/* 
  Every node of the infinite tree has an integer index.
  The root node has index 1.  If $k$ is the index of a node, 
  then $2*k$ and $2*k+1$ are the indices of its children. */
  
nat8 dyg_depth(dyg_node_index k);
/*
  The depth of a node with index $k$.  The root's depth is 1. */
  
#define dyg_MAX_INDEX (1152921504606846975L)
#define dyg_MAX_DEPTH(d) (60/d)
/*
  The maximum node index is $2^60-1$, meaning that in any concrete
  tree the maximum depth of an existing node is 60 (counting the depth
  of the root as 1).  
  
  Thus the finest achievable 4-dimensional (3D+time) grid has
  $2^15 = 32768$ cells in each direction. That is enough for 1m
  resolution in a 30 km wide 3-d field, and 10 min resolution in a
  200-day simulation period.
  
  The finest achievable 3-dimensional grid has $2^20 = 1048576$ cells
  in each direction, which means 10 cm resolution in a 100 km wide
  field, and 1 min resolution in a 700-day simulation period. */

/* THE TREE STRUCTURE */

/* It is not obvious whether the tree structure must be stored at all. */

typedef struct dyg_node {
  struct dyg_node  *p;   /* Parent node */
  struct dyg_node  *c0;  /* "Low" children node */
  struct dyg_node  *c1;  /* "High" children node */
} dyg_node;

typedef struct dyg_header {
  nat8  d;         /* Dimension */
  dyg_node *root;  /* Root node of the tree */
} dyg_header;

dyg_header *dyg_new_tree(nat8 d);
/*
  Creates a new $d$-dimensional tree with a single node (the root). */
  
dyg_node *dyg_new_node(dyg_node *p);
/*
  Creates a new node with no children and parent $p$. */
  
void dyg_free_node(dyg_node **pp);
/*
  Reclaims the storage used by node $*pp$, which must not have
  any children, and sets $(*pp) = NULL$. */
  
void dyg_split_node(dyg_node *p);
/*
  The node $p$ must have no children.  Creates two new nodes and
  appends them as children of $p$. */
  
void dyg_unsplit_node(dyg_node *p);
/*
  The node $p$ must have two children but no grandchildren.
  Deletes and reclaims the two children of $p$. */

#endif
