/* N-pulses - Compact unit-partition finite elements. */
/* Last edited on 2005-04-22 10:12:14 by stolfi */

#ifndef dg_pulse_N_H
#define dg_pulse_N_H

#include <dg_grid.h>
#include <bz_basic.h>
#include <dg_pulse.h>
#include <js.h>

/*
  N-PULSES
  
  The dyadic N-pulses are defined for any {c \geq -1} and any
  non-negative degree {g \geq 2*c+1}. 
  
  For a given {g} and {c}, there are {g-c} distinct mother pulses
  {w^g_c[m]}. They fall into two sub-classes:

  * When {m} is in {c+1..g-c-1}, the mother spline has
    support {(0_1)}, and coincides with the Bernstein-Bzier
    polynomial {BB^g_m}:

      { w^g_c[m](z) = BB^g_m(z) = choose(g,m) z^{g-m} (1-z)^m }

  * When {m} is in {0..c}, the mother spline is supported on 
    two grid cells, spanning {(0_2)}. In the first cell {(0_1)},
    it can be expressed as a combination of Bernstein-Bzier
    polynomials:
    
      { w^g_c[m](z) = SUM { choose(k,m)/2^{k} BB^g_{g-c+k}(z) : k = m..c } }
    
    For {z} in {(1_2)}, the following relation holds

      { w^g_c[m](z) = w^g_c[c-m](2-z) }
      
    so 
      
      { w^g_c[m](z) = SUM { choose(c-k,m)/2^{c-k} BB^g_{g-k}(z) : k = 0..c-m }
        
  */

unsigned int dg_pulse_N_num_mothers (dg_cont_t c, dg_degree_t g);
  /* Returns the number of mother N-pulses with continuity {c}
    and degree {g}. Returns 0 if such N-pulses aren't defined or 
    aren't supported by this module. */

dg_grid_size_t dg_pulse_N_mother_supp_count
  ( dg_cont_t c, 
    dg_degree_t g, 
    dg_pulse_mother_index_t m
  );
  /* Returns the number of unit-grid cells that comprise the support
    of the mother N-pulse with given {c,g,m}. Fails if such N-pulses
    aren't defined or aren't supported by this module. */
    
void dg_pulse_N_mother_to_bezier
  ( dg_cont_t c, 
    dg_degree_t g, 
    dg_pulse_mother_index_t m,
    dg_grid_pos_t x,
    double *bz
  );
  /* Stores into {bz[0..g]} the Bzier coefficients of the mother
    N-pulse with continuity {c}, degree {g}, and mother index {m}, for
    the interval {(x _ x+1)}. Fails if such pulse is not defined or
    supported by this module. The coefficients will be non-zero only
    if {x == 0}, or if {x == 1} and {m <= c}. */

#endif
