/* Finite elements for multidimensional dyadic spline spaces. */
/* Last edited on 2005-06-05 15:28:29 by stolfi */

#ifndef dg_tent_H
#define dg_tent_H

#include <dg_grid.h>
#include <bz_basic.h>
#include <dg_pulse.h>
#include <bz_patch.h>
#include <vec.h>
#include <bool.h>

/*
  STANDARD DYADIC TENTS
  
  For multidimensional grids, the finite-element bases that we
  consider here consist of /standard dyadic finite tents/, or /dyadic
  tents/ for short.
  
  Each dyadic tent is a polynomial spline defined on some level {r} of
  the {d}-dimensional dyadic multigrid {G*}. Indeed, a standard dyadic
  tent is the product of {d} standard dyadic pulses, each depending on
  only one coordinate:
  
    { t(x) = PROD { b[i](x[i]) : i \in 0..d-1 } }
  
  Each of these pulses has grid size {gsz[i] = dg_grid_size(d,r,i)}.
  
  TENT SUPPORT
  
  It folows that the support of a standard dyadic tent of rank {r} is
  a cell pack of rank {r}, which is the "Cartesian product" of the
  supports of its factor pulses {b[0..d-1]}. The support may wrap
  around the boundary of {D}, and may overlap itself if {r} is small
  enough.
  
  From the conventions on standard pulse supports, the lowest cell of
  {C} of the pack (the tent's /anchor cell/) is the cell with
  position vector {pos[0..d-1]} in level {r} of {G*}.  The number of
  cells {n[i]} in the tent's support along axis {i} can be determined by
  
    {n[i] = dg_pulse_supp_count(pkind[i],c[i],g[i],m[i],gsz[i])}
    
  where {gsz[i]} is the grid size of level {r} of {G*} along direction {i}.
  The tent's support {B} then spans intervals {pos[i]..pos[i]+n[i]-1} along
  axis {i}, modulo {gsz[i]}.

  TENT EVALUATION
  
  The following procedures compute the value and partial derivatives of
  a dyadic tent {t} at a point {x} of {R^d}. They properly handle
  wrap-around and self-overlapping of the mother pulses.
  
  The results are stored in {f[0..N-1]}, for some {N}: the function
  value is stored at {f[0]}, followed by the other derivatives, as
  requested.
  
  The root cell {R} defaults to {(0_1)^d} if {R == NULL}.
*/

void dg_tent_eval
  ( dg_dim_t d,
    dg_pulse_family_t fam[], 
    dg_pulse_t p[],
    interval_t R[],
    double x[],
    dg_degree_t ord[],
    double f[]
  );
  /* Evaluates a tent function and all its partial derivatives up to 
    order {ord} at a point {x[0..d-1]}.
    
    Computes all derivatives up to order {ord[i]} in each
    coordinate {x[i]}. The derivatives are stored in {f[0..N-1]},
    where  {N = PROD { ord[i]+1 : i = 0..d-1 } },
    in ``column by column'' order. E.g., for {d=3,ord=(2,1,2)},
    the {f} vector will contain the {3*2*3 = 18} derivatives
        
      { (w,wx,wxx,wy,wxy,wxxy,wz,wxz,wxxz,wyz,wxyz.wxxyz,wzz, ... wxxyzz) }
      
    where {wxxy} means {w} differentiated twice on {x[1]} and once on {x[2]}. */

void dg_tent_eval_total
  ( dg_dim_t d,
    dg_pulse_family_t fam[], 
    dg_pulse_t p[],
    interval_t R[],
    double x[],
    dg_degree_t ord,
    double f[]
  );
  /* Same as {dg_tent_eval}, but computes only derivatives up to 
    *total* order {ord} in each coordinate {x[i]}. The derivatives are
    stored in {f[0..N-1]}, where {N = choose(ord+d,d) }, sorted
    by total order and then lexicographically. E.g., for {d=3,ord=2},
    the {f} vector will contain the {choose(5,3) = 10} derivatives

       { (w,wx,wy,wz,wxx,wxy,wyy,wxz,wyz,wzz) }.

  */

/*
  MOTHER TENT INDEX
  
  A /mother tent index/ {tm} is a single integer that conveniently
  combines {d} mother pulse indices {m[0..d-1]}, in a way that depends
  on the family vector {fam[0..d-1]}.  The indices are combined by
  interleaving their bits in cyclic order, so that bit {j} of {m[i]}
  is bit {j*d + i} of {tm}.
  
  It follows that a tent can be specified completely by two integers:
  its mother tent index {tm}, and the cell index {ck} of the lower
  cell of its support.  */
  
typedef uint32_t dg_tent_mother_index_t;
  /* A packed vector of {dg_pulse_mother_index_t}s. */

dg_tent_mother_index_t dg_tent_mother_pack(dg_dim_t d, dg_pulse_mother_index_t m[]);
  /* Packs pulse mother indices {m[0..d-1]} into a tent mother index.  */

void dg_tent_mother_unpack(dg_dim_t d, dg_tent_mother_index_t tm, dg_pulse_mother_index_t m[]);
  /* Unpacks a tent mother index {tm} into pulse mother indices {m[0..d-1]}.  */

typedef struct dg_tent_t  /* Identifies a tent within a tent family. */
  { dg_cell_index_t ck;    /* Index of anchor cell. */
    dg_tent_mother_index_t tm;   /* Packed mother pulse indices. */
  } dg_tent_t;
  /* Identifies (within some family fixed by context) the tent whose
    mother tent index is {tm} and whose anchor cell has cell index
    {ck}.  */

dg_tent_t dg_tent_pack
  ( dg_dim_t d, 
    dg_pulse_family_t fam[],
    dg_pulse_t p[]
  );
  /* Returns the packed description of the tent formed by pulses
    {p[0..d-1]} from families {fam[0..d-1]}, respectively. */

void dg_tent_unpack
  ( dg_dim_t d, 
    dg_pulse_family_t fam[],
    dg_tent_t *t,
    dg_pulse_t p[]
  );
  /* Unpacks a packed tent description {t} into its factor pulses
    {p[0..d-1]}, assuming they belong to families {fam[0..d-1]}. */

/* 
  ENUMERATING TENTS */

typedef struct dg_tent_vec_t { unsigned int nel; dg_tent_t *el; } dg_tent_vec_t;
  /* A list of {dg_tent}s. */

dg_tent_vec_t dg_tents_from_cells
  ( dg_dim_t d, 
    dg_tent_mother_index_t tm,
    dg_cell_index_vec_t ck
  );
  /* Returns a list of {dg_tent_t} {t} obtained by combining tent
    index {tm} with all cell indices {ck[0..]} */

/* PRINTOUT: */
  
void dg_tent_print
  ( FILE *wr, 
    dg_dim_t d,
    dg_pulse_family_t fam[], 
    dg_pulse_t p[]
  );
  /* Prints the tent whose factors are pulses {p[0..d-1]} from
    families {fam[0..d-1]}, in the format "{pulse[0]}*...{pulse[d-1]}"
    where {pulse[i]} is the output of {dg_pulse_print}. */

void dg_tent_packed_print
  ( FILE *wr, 
    dg_dim_t d,
    dg_pulse_family_t fam[], 
    dg_tent_t *t
  );
  /* Like {dg_tent_print} but extracts the tent's info
    from the packed tent {t}. */

/* 
  MANIPULATING VECTORS OF {dg_tent_t}: */

dg_tent_vec_t dg_tent_vec_new(unsigned int nel);

#define dg_tent_vec_expand(nv,index) \
  vec_expand(vec_cast_ref(nv), index, sizeof(dg_tent_t))
  /* Makes sure that the vector {nv} has element {nv.el[index]}.
    Reallocates {nv.el} if necessary, to have at least {index+1}
    elements -- but possibly more. */

#define dg_tent_vec_trim(nv,nel) \
  vec_trim(vec_cast_ref(nv), nel, sizeof(dg_tent_t))
  /* Makes sure that the vector {nv} has exactly {nel} elements.
    Reallocates {nv.el} if necessary. */

#endif
